"""
Module for helper functions in Orderly SDK
"""

import asyncio


def get_loop():
    """check if there is an event loop in the current thread, if not create one
    inspired by https://stackoverflow.com/questions/46727787/runtimeerror-there-is-no-current-event-loop-in-thread-in-async-apscheduler
    """
    try:
        loop = asyncio.get_event_loop()
        return loop
    except RuntimeError as e:
        if str(e).startswith("There is no current event loop in thread"):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return loop
        else:
            raise



import asyncio
import base64
import datetime
import json as jsonlib
import logging
from collections import defaultdict
from typing import DefaultDict, Dict, Optional

import base58
import websockets
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from websockets import WebSocketClientProtocol

from .helpers import get_loop

logging.basicConfig(
    format="%(asctime)s %(levelname)s %(name)s %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


class WsTopicManager:
    endpoint: str
    queues: DefaultDict[str, asyncio.Queue]
    websocket: WebSocketClientProtocol

    def __init__(
        self,
        _id="WS_PUBLIC",
        account_id="",
        endpoint="",
        loop=None,
    ):
        self._id = _id
        self.account_id = account_id
        self.endpoint = endpoint + self.account_id
        self.loop = loop or get_loop()
        # topic -> topic event queue
        self.queues: DefaultDict[str, asyncio.Queue] = defaultdict(asyncio.Queue)

    async def _connect(
        self,
        timeout: Optional[int | float] = None,
        **kwargs,
    ):
        async for websocket in websockets.connect(self.endpoint, **kwargs):
            try:
                self.websocket = websocket
                await self._reconnect()
                logger.debug(f"Connected to {self.endpoint}")
                while True:
                    try:
                        message = await asyncio.wait_for(
                            websocket.recv(), timeout=timeout
                        )
                        await self._handle_heartbeat(message)
                        # asyncio.create_task(self._handle_message(message))
                        await self._handle_message(message)
                    except asyncio.TimeoutError:
                        logger.warning(f"Connection to {self.endpoint} timed out")
                        break
            except websockets.ConnectionClosed:
                logger.warning(f"Disconnected from {self.endpoint}")
            except Exception as e:
                logger.exception(e)

    def start(self, timeout: Optional[int | float] = None, **kwargs):
        self.loop.call_soon_threadsafe(
            asyncio.create_task, self._connect(timeout, **kwargs)
        )

    def subscribe(self, topic):
        self.queues[topic] = asyncio.Queue()

    async def do_subscribe(self, topic):
        await self.send_json({"id": self._id, "event": "subscribe", "topic": topic})

    async def request(self, symbol: str):
        params = {"params": {"type": "orderbook", "symbol": symbol}}
        await self.send_json({"id": self._id, "event": "request", **params})

    async def unsubscribe(self, topic):
        await self.websocket.send(
            jsonlib.dumps({"id": self._id, "event": "unsubscribe", "topic": topic})
        )
        self.queues.pop(topic)

    async def send_json(self, message):
        if "event" in message and message["event"] != "pong":
            logger.debug(f"sending message to {self.endpoint}: {message}")
        await self.websocket.send(jsonlib.dumps(message))

    async def _handle_heartbeat(self, message):
        message = jsonlib.loads(message)
        if "event" in message:
            if message["event"] == "ping":
                await self.send_json({"event": "pong"})

    async def _handle_request_orderbook(self, message: Dict):
        topic = message["data"]["symbol"] + "@orderbook"
        data = message["data"]
        data["ts"] = message["ts"]
        await self.queues[topic].put(data)

    async def _handle_general_message(self, message: Dict):
        data = message["data"]
        # data["ts"] = message["ts"]
        # logger.info(f"received message from {self.endpoint}: {message}")
        await self.queues[message["topic"]].put(data)

    async def recv(self, topic, timeout=10):
        res = None
        while not res:
            try:
                res = await asyncio.wait_for(self.queues[topic].get(), timeout=timeout)
            except asyncio.TimeoutError:
                logging.info(f"no message in {timeout} seconds")
        return res

    async def _handle_message(self, message):
        message = jsonlib.loads(message)
        # logger.info(f"received message from {self.endpoint}: {message}")
        if "event" in message:
            if message["event"] not in ("ping", "pong"):
                if message["success"]:
                    if "data" in message and message["event"] == "request":
                        await self._handle_request_orderbook(message)
                    return
                else:
                    raise Exception(message)
        if "data" in message:
            await self._handle_general_message(message)

    # reconnect, resubscribe
    async def _reconnect(self):
        for topic in self.queues.keys():
            await self.do_subscribe(topic)


class OrderlyPublicWsManager(WsTopicManager):
    def __init__(
        self,
        _id="WS_PUBLIC",
        account_id="",
        endpoint="",
        loop=None,
    ):
        super().__init__(
            _id=_id,
            account_id=account_id,
            endpoint=endpoint,
            loop=loop,
        )


class OrderlyPrivateWsManager(WsTopicManager):
    def __init__(
        self,
        _id="WS_PRIVATE",
        account_id="",
        orderly_key: Optional[str] = None,
        orderly_secret: Optional[str] = None,
        endpoint="",
        loop=None,
    ):
        super().__init__(
            _id=_id,
            account_id=account_id,
            endpoint=endpoint,
            loop=loop,
        )
        self.orderly_key = orderly_key
        self.orderly_secret = orderly_secret
        if orderly_secret is not None:
            self.orderly_private_key = Ed25519PrivateKey.from_private_bytes(
                base58.b58decode(orderly_secret)[0:32]
            )

    def _signature(self, data):
        data_bytes = bytes(str(data), "utf-8")
        request_signature = base64.b64encode(
            self.orderly_private_key.sign(data_bytes)
        ).decode("utf-8")
        return request_signature

    async def _login(self) -> None:
        ts = round(datetime.datetime.now().timestamp() * 1000)
        await self.send_json(
            {
                "id": self._id,
                "event": "auth",
                "params": {
                    "orderly_key": f"ed25519:{self.orderly_key}",
                    "sign": self._signature(ts),
                    "timestamp": str(ts),
                },
            }
        )

    # reconnect, resubscribe
    async def _reconnect(self):
        for topic in self.queues.keys():
            await self._login()
            await self.do_subscribe(topic)



"""
Module for Orderly API exceptions
"""


class OrderlyAPIException(Exception):
    """
    Common exception for Orderly API
    """

    def __init__(self, resp_json, status_code):
        self.status_code = status_code
        try:
            self.code = resp_json.get("code")
            self.message = resp_json.get("message")
        except Exception:
            self.code = None
            self.message = resp_json

    def __str__(self):
        return f"APIError(code={self.code}: {self.message}"


class OrderlyRequestException(Exception):
    """
    Value exception for Orderly API
    """

    def __init__(self, message) -> None:
        self.message = message

    def __str__(self) -> str:
        return f"OrderlyRequestException: {self.message}"



import base64
import datetime
import json as jsonlib
import logging
from collections import defaultdict
from typing import Dict, Optional
from urllib.parse import urlparse

import aiohttp
import base58
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey

from .exceptions import OrderlyAPIException, OrderlyRequestException
from .helpers import get_loop

logging.basicConfig(
    format="%(asctime)s %(levelname)s %(name)s %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


class AsyncClient:
    _id: str
    account_id: Optional[str]
    orderly_key: Optional[str]
    orderly_secret: Optional[str]
    endpoint: Optional[str]
    api_version: str = "v1"
    timeout = 30

    def __init__(
        self,
        _id="websocket_client",
        account_id: Optional[str] = None,
        orderly_key: Optional[str] = None,
        orderly_secret: Optional[str] = None,
        endpoint: Optional[str] = None,
        loop=None,
    ):
        self._id = _id
        self.account_id = account_id
        self.orderly_key = orderly_key
        self.orderly_secret = orderly_secret
        if orderly_secret is not None:
            self.orderly_private_key = Ed25519PrivateKey.from_private_bytes(
                base58.b58decode(orderly_secret)[0:32]
            )
        self.endpoint = endpoint
        self.loop = loop or get_loop()
        self.session = self._init_session()
        self.headers: Dict[str, str] = defaultdict(str)
        self.response = None

    def _get_headers(self) -> Dict:
        headers = {
            "Accept": "application/json",
        }
        if self.orderly_key:
            headers["orderly-key"] = self.orderly_key
        return headers

    def _init_session(self) -> aiohttp.ClientSession:
        return aiohttp.ClientSession(loop=self.loop, headers=self._get_headers())

    async def close_connection(self):
        if self.session:
            assert self.session
            await self.session.close()

    async def _request(
        self,
        method,
        uri: str,
        signed: bool,
        params: Optional[Dict],
        json: Optional[Dict],
    ):
        if signed:
            ts = round(datetime.datetime.now().timestamp() * 1000)
            json_str = ""
            if json is not None:
                logger.debug(f"request json body: {json}")
                json_str = jsonlib.dumps(json)
            path = urlparse(uri).path
            # logger.info(f"request path: {path.path}")
            signature_str = f"{ts}{method.upper()}{path}{json_str}"
            logger.debug(f"request signature: {signature_str}")
            data_bytes = bytes(signature_str, "utf-8")

            req_signature = base64.b64encode(
                self.orderly_private_key.sign(data_bytes)
            ).decode("utf-8")

            self.headers["orderly-signature"] = req_signature
            if self.account_id is not None:
                self.headers["orderly-account-id"] = self.account_id
            else:
                pass
            self.headers["orderly-key"] = f"ed25519:{self.orderly_key}"
            self.headers["orderly-timestamp"] = str(ts)
            self.headers["Content-Type"] = (
                "application/json" if json else "application/x-www-form-urlencoded"
            )
            self.headers["Cache-Control"] = "no-cache"
            self.session.headers.update(self.headers)

        logger.debug("request uri: %s", uri)
        async with getattr(self.session, method)(
            uri, params=params, json=json
        ) as response:
            self.response = response
            return await self._handle_response(response)

    async def _handle_response(self, response: aiohttp.ClientResponse):
        if not str(response.status).startswith("2"):
            logger.error("response: %s", response)
            raise OrderlyAPIException(response, response.status)
        try:
            return await response.json()
        except ValueError as exc:
            txt = await response.text()
            raise OrderlyRequestException(f"Invalid Response: {txt}") from exc

    def _create_rest_uri(self, ep: str, v: str = ""):
        if not v:
            v = self.api_version
        return f"{self.endpoint}/{v}/{ep}"

    async def _request_api(
        self, method, ep: str, signed: bool, v: str = "", params=None, json=None
    ):
        uri = self._create_rest_uri(ep, v)
        return await self._request(method, uri, signed, params=params, json=json)

    async def _get(self, ep, signed=False, v: str = "", params=None, json=None):
        return await self._request_api("get", ep, signed, v, params, json)

    async def _post(self, ep, signed=False, v: str = "", params=None, json=None):
        return await self._request_api("post", ep, signed, v, params, json)

    async def _put(self, ep, signed=False, v: str = "", params=None, json=None):
        return await self._request_api("put", ep, signed, v, params, json)

    async def _delete(self, ep, signed=False, v: str = "", params=None, json=None):
        return await self._request_api("delete", ep, signed, v, params, json)

    async def get_maintenance_info(self) -> Dict:
        return await self._get("public/system_info")

    async def get_user_statistics(self) -> Dict:
        return await self._get("client/statistics", True)

    async def create_order(self, json: Dict) -> Dict:
        return await self._post("/order", True, json=json)

    async def claim_liquidated_positions(self, json: Dict) -> Dict:
        return await self._post("/liquidation", True, json=json)

    async def claim_insurance_fund(self, json: Dict) -> Dict:
        return await self._post("/claim_insurance_fund", True, json=json)

    async def get_all_positions(self) -> Dict:
        return await self._get("/positions", True)
